/*
   ReadIMX.dll
   (c) LaVision
	TL 10.05.01: reading IMG/IMX/VEC raw data
   TL 11.06.01: added reading of string attributes
	TL 30.09.02: corrected mis-aligned image_header for Linux
	TL 05.02.04: support IM7/VC7 on WIN32
	TL 22.03.04: BufferScaleType and reading of scaling information

   Read a LaVision IMX/IMG/VEC file by a call to:

   extern "C" int EXPORT ReadIMX ( const char* theFileName, BufferType* myBuffer, AttributeList** myList )

   If you don't want to read the attributes, set myList = NULL. If myList is set,
   the parameter returns a pointer to the list. You have to free the strings manually!

   The function returns the error codes of ImReadError_t.
*/

#include "ReadIMX.h"


#ifdef _WIN32

#	ifdef __BORLANDC__
#		pragma option -a1
#	else
#		pragma pack (1)
#	endif

#endif


typedef struct		// image file header (256 Bytes)
	{
	 short int 	imagetype;     // 0:  (Image_t)
	 short int	xstart;			// 2:  start-pos left
	 short int	ystart;			// 4:  start-pos top
	 Byte			extended[4];	// 6:  reserved
	 short int	rows;				// 10: total number of rows, if -1 the real number is stored in longRows
	 short int	columns;			// 12: total number of columns, if -1 see longColumns
	 short int  image_sub_type;// 14: type of image (int):
										//				(included starting from sc_version 4.2 on)
										// 	0 = normal image
										//		1 = PIV vector field with header and 4*2D field
										//				rows = 9 * y-size
										//		2 = simple 2D vector field (not used yet)
										//				rows = 2 * y-size
										//		3 ...
	 short int 	y_dim;			// 16: size of y-dimension (size of x-dimension is always = columns)
	 short int 	f_dim;			// 18: size of f-dimension (number of frames)

	 // for image_sub_type 1/2 only:
	 short int	vector_grid; 	// 20: 1-n = vector data: grid size
										//     		(included starting from sc_version 4.2 on)

	 char			ext[11];			// 22: reserved

	 Byte			version;			// 33:  e.g. 120 = 1.2	300+ = 3.xx  40-99 = 4.0 bis 9.9
#ifdef _WIN32
	 char       date[9];       // 34	
	 char       time[9];			// 43
#else
//#message MIS-ALIGNED!!!
	 char       date[8];       // 34	
	 char       time[8];			// 43
#endif
	 short int  xinit;         // 52:  x-scale values
    float      xa;            // 54
    float      xb;            // 58
    char       xdim[11];      // 62
    char       xunits[11];    // 73
	 short int  yinit;         // 84:  y-scale values
    float      ya;            // 86
    float      yb;            // 90
    char       ydim[11];      // 94
    char       yunits[11];    // 105
	 short int  iinit;         // 116: i-scale (intensity) values
    float      ia;            // 118
    float      ib;            // 122
    char       idim[11];      // 126
    char       iunits[11];    // 137

	 char       com1[40];      // 148: comment (2 lines)
	 char       com2[40];		// 188

	 long			longRows;		// 228: (large) number of rows, TL 04.02.2000
	 long			longColumns;	// 232: (large) number of columns, TL 04.02.2000
    long			longZDim;		// 236: (large) number of z-dimension, TL 02.05.00

	 char			reserved[12];	// 240: reserved
	 long    	checksum;		// 252-255
	} image_header;


#ifdef __BORLANDC__
#	pragma option -a.
#else
#	pragma pack ()
#endif


enum	Image_t
	{	IMAGE_IMG 			= 18,	 	// uncompressed WORD image
		IMAGE_IMX,						// compressed WORD image
		IMAGE_FLOAT,            	// floating point uncompressed
		IMAGE_SPARSE_WORD,			// sparse buffer with word scalars
		IMAGE_SPARSE_FLOAT,			// sparse buffer with float scalars
	};

#define VER_EXTHEADER			54
#define VER_VOLUME_BUFFER		55


void SetBufferScale( BufferScaleType* theScale, float theFactor, float theOffset, const char* theDesc, const char* theUnit )
{
	if (theScale)
	{
		theScale->factor	= theFactor;
		theScale->offset	= theOffset;
		strncpy( theScale->description, theDesc, sizeof(theScale->description) );
		theScale->description[sizeof(theScale->description)-1] = 0;
		strncpy( theScale->unit, theUnit, sizeof(theScale->unit) );
		theScale->unit[sizeof(theScale->unit)-1] = 0;
	}
}


int CreateBuffer( BufferType* myBuffer, int theNX, int theNY, int theNZ, int theNF, int isFloat, int vectorGrid, int imageSubType )
{
	if (myBuffer==NULL)
		return 0;
   myBuffer->isFloat = isFloat;
   myBuffer->nx = theNX;
   myBuffer->ny = theNY;
   myBuffer->nz = theNZ;
   myBuffer->nf = theNF;
   myBuffer->totalLines = theNY*theNZ*theNF;
	myBuffer->vectorGrid = vectorGrid;
	myBuffer->image_sub_type = imageSubType;
   int size = theNX * myBuffer->totalLines;
   if (isFloat)
	{
      myBuffer->floatArray = (float*)malloc(sizeof(float)*size);
   }
	else
	{
      myBuffer->wordArray = (Word*)malloc(sizeof(Word)*size);
   }
	SetBufferScale( &myBuffer->scaleX, 1, 0, "", "pixel" );
	SetBufferScale( &myBuffer->scaleY, 1, 0, "", "pixel" );
	SetBufferScale( &myBuffer->scaleI, 1, 0, "", "counts" );
   return (myBuffer->floatArray!=NULL || myBuffer->wordArray!=NULL);
}


Byte* Buffer_GetRowAddrAndSize( BufferType* myBuffer, int theRow, unsigned long &theRowLength )
{
	theRowLength = myBuffer->nx * (myBuffer->isFloat?sizeof(float):sizeof(Word));
	if (myBuffer->isFloat)
		return (Byte*)&myBuffer->floatArray[myBuffer->nx*theRow];
	return    (Byte*)&myBuffer->wordArray [myBuffer->nx*theRow];
}


const size_t PAGESIZE = 0x800;		// Words !  size must not be too big!

long IMX_pagepos = 0;   // TL 16.12.1999

// read data from the image file to the buffer, reset 2 variables
// called by SCBuffer:Read
Word ReadNextPage (FILE* file, Word* page, signed char** bpageadr, Word* count)
{
	IMX_pagepos = ftell(file);
	if (fread(page, 1, PAGESIZE, file)==0)
		return 0;			// read error? (not TRUE if EOF)
	else
	{	*count = 0;
		*bpageadr = (signed char*) page;
		return (Word)(ftell(file) - IMX_pagepos);
	}
}


ImReadError_t SCPackOldIMX_Read( FILE* theFile, BufferType* myBuffer )
{
	enum {BIT8,			// compressed as 8 bit
			BIT4L,		// compressed to 4 bit left nibble
			BIT4R,		// compressed to 4 bit right nibble
			BIT16 }		// n-uncompressed 2-byte pixels
			cmode;		// compression mode
	/*
	compression syntax:
	  BIT8: - At start lastvalue = 0 and cmode = BIT8, i.e. it is assumed
		  that the following bytes are coded with 1 pixel per byte
	  BIT8: - If -128 is encountered, a pixel as a Word (2 bytes) follows.
		  This value is the absolute value of the pixel.
	  BIT8: - If -127 is encountered cmode is set to BIT4, i.e. from now
		  on, each byte contains 2 pixels, each consisting of 4 bits.
		  Thus each new value can take on values between -7 and +7.
		  This is coded as 0x00-0x07 (0-7) and 0x09-0x0F (-7 to -1)
	  BIT8: - If 127 is encountered, then the next byte gives the number
		  of uncompressed 16-bit pixel values, followed by the 2-byte
		  data.
	  BIT4: - Cmode BIT4 is set back to BIT8 if 0x08 is encountered in the
		  right 4 bits.
	*/

	// Yes. Skip the preview image stored after the header
	Word bytecount, count, lastvalue, newvalue = 0;
	signed char *bpageadr, bvalue = 0, newnibble;
	Byte nx, ny, nbytes = 0;
	int bline;
	Word  poge [PAGESIZE];							// input buffer to increase performance
	Word* page = (Word*)&poge;						// by reducing the # of read operations

	if (!fread(&nx,1,1,theFile)) 	goto errexit;	   // read x-size of preview
	if (!fread(&ny,1,1,theFile))  goto errexit;		// read y-size of preview
	if (fseek( theFile, nx*ny, SEEK_CUR))  goto errexit;	// read preview

	cmode = BIT8;                               // start with BIT8 mode
	lastvalue = 0;                              // and assuming a previous 0
	if ( (bytecount = ReadNextPage(theFile, page, &bpageadr, &count)) == 0 )  goto errexit;

	for (bline = 0; bline < myBuffer->totalLines; bline++ )
	{
		Word* rowW = &myBuffer->wordArray[myBuffer->nx * bline];
		for (int i = 0; i < myBuffer->nx; i++ )
		{
			if ( count == bytecount )               // no more bytes left?
				if ( (bytecount = ReadNextPage(theFile, page, &bpageadr, &count)) == 0 )
					goto errexit;	// error 5
			if ( cmode == BIT4L || cmode == BIT8 )  // need new byte
			{
				bvalue = *bpageadr++;                 // new value
				count++;                              // 1 byte processed
			}

			if ( cmode == BIT4R )                   // process right nibble next
			{
				newnibble = bvalue & 0x0F;            // get right nibble
				cmode = BIT4L;                        // next: left nibble
				if ( newnibble == 0x08 )              // change back to BIT8
				{
					cmode = BIT8;
					if ( count == bytecount )           // no more bytes left?
						if ( (bytecount = ReadNextPage(theFile, page, &bpageadr, &count)) == 0 )
							goto errexit;
					bvalue = *bpageadr++;               // new value BIT8-mode
					count++;                            // 1 byte processed
				}
				else
				{
					if ( newnibble & 0x08 ) newnibble |= 0xF0;  // -1 to -7
					newvalue = lastvalue + (long)newnibble;
				}
			}
			else if ( cmode == BIT4L )              // process left nibble next
			{
				newnibble = bvalue >> 4;
				if ( newnibble & 0x08 ) newnibble |= 0xF0;    // -1 to -7
				newvalue = lastvalue + (long)newnibble;       // get left nibble
				cmode = BIT4R;                        // next: right nibble
			}

			if ( cmode == BIT8 )                    // BIT8 compression mode
			{
				switch ( bvalue )
				{
				case -128:                     // exception?
					cmode = BIT16;              // change to BIT16
					nbytes = 1;                 // read only 1 pixel
					break;
				case 127:                      // exception? n-16-bit values; number of bytes follows
					if ( count == bytecount )         // no more bytes left?
						if ( (bytecount = ReadNextPage(theFile, page, &bpageadr, &count)) == 0 )
							goto errexit;
					cmode = BIT16;                    // set cmode correctly
					nbytes = *( (Byte*)bpageadr++ );  // number of bytes
					count++;                          // 1 byte read
					break;
				case -127:                          // change to BIT4 mode
					if ( count == bytecount )         // no more bytes left?
						if ( (bytecount = ReadNextPage(theFile, page, &bpageadr, &count)) == 0 )
							goto errexit;
					bvalue = *bpageadr++;             // get new byte
					count++;
					newnibble = bvalue >> 4;
					if ( newnibble & 0x08 ) newnibble |= 0xF0; // -1 to -7
					newvalue = lastvalue + (long)newnibble;    // get left nibble
					cmode = BIT4R;                    // process right nibble next
					break;
				default:
					newvalue = lastvalue + (long)bvalue;  // get new value, normal mode
				}
			}			// if cmode = BIT8

			if ( cmode == BIT16 )                   // nbytes 16-bit pixels
			{
				if ( count == bytecount )             // no more bytes left?
					if ( (bytecount = ReadNextPage(theFile, page, &bpageadr,	&count)) == 0 )
						goto errexit;		// error 5
				if ( count == bytecount-1 )           // only one byte left?
				{                                     // ---> trouble!
					newvalue = (Word)*(Byte*)bpageadr;  // get low byte first from old page
					if ( (bytecount = ReadNextPage(theFile, page, &bpageadr,	&count)) == 0 )
						goto errexit;		// error 5
					newvalue += ((Word)*(Byte*)bpageadr++) << 8; // get high byte from new page
					count = 1;                          // one byte used already
				}
				else
				{
					newvalue = *( (Word*)bpageadr );    // get new value, Word pointer!
					bpageadr += 2;                      // increment byte pointer
					count += 2;                         // 2 bytes processed
				}
				if ( --nbytes == 0 ) cmode = BIT8;    // decrement counter, change mode
			}

			if ( i < myBuffer->nx )                     // maybe skip right part
				rowW[i] = newvalue;
				//pixbuf.W[bline][i] = newvalue;			// store in pixbuf
			lastvalue = newvalue;                  // update last value

		} 	// for i
	}	// for bline

	fseek( theFile, IMX_pagepos + ((char*)bpageadr - (char*)page), SEEK_SET );
	return IMREAD_ERR_NO;

errexit:
	return IMREAD_ERR_DATA;
}


extern "C" int EXPORT ReadIMX ( const char* theFileName, BufferType* myBuffer, AttributeList** myList )
{
	image_header header;

	FILE* theFile = fopen(theFileName, "rb");				// open for binary read
	if (theFile==NULL)
	   return IMREAD_ERR_FILEOPEN;

	// Read an image in our own IMX or IMG or VEC or VOL format
	int theNX,theNY,theNZ,theNF;
	// read and store file header contents
	if (!fread ((char*)&header, sizeof(header), 1, theFile))
	{
      fclose(theFile);
      return IMREAD_ERR_HEADER;
   }
   
	int itsVersion	= header.version;
	//itsDate		= header.date;
	//itsTime		= header.time;

   // scales
   char olddim[12], oldunit[12];
   strncpy(olddim,header.xdim,11); olddim[11]=0;
   strncpy(oldunit,header.xunits,11); oldunit[11]=0;
	SetBufferScale( &myBuffer->scaleX, header.xa, header.xb, olddim, oldunit );
   strncpy(olddim,header.ydim,11); olddim[11]=0;
   strncpy(oldunit,header.yunits,11); oldunit[11]=0;
	SetBufferScale( &myBuffer->scaleY, header.ya, header.yb, olddim, oldunit );
   strncpy(olddim,header.idim,11); olddim[11]=0;
   strncpy(oldunit,header.iunits,11); oldunit[11]=0;
	SetBufferScale( &myBuffer->scaleI, header.ia, header.ib, olddim, oldunit );

	// copy the comments. Before copying them to CStrings,
	// add zeros after the last characters in case it has the max. length (40)
   /*
	c1 = header.com1[40], c2 = header.com2[40];		// (= access out of bounds)
	header.com1[40] = 0; itsComment = header.com1; header.com1[40] = c1;
	header.com2[40] = 0;
   if (strlen(header.com2)>0) {
      itsComment += "\n";
      itsComment += header.com2;
   }
   header.com2[40] = c2;
   */

	if (itsVersion > 99)		// old versions; the new numbers lie between 40 and 99
	{	// make old data file formats compatible with version 3.0 and higher
		if ( itsVersion < 210 )								// = version 2.1
		{	header.columns	= 384;	                  // image size
			header.rows		= 286;
			header.xstart  = header.ystart  = 0;		// image shift
			header.com1[0] = header.com2[0] = 0;		// comment
		}
		//if ( itsVersion < 120 )								// = version 1.2
		//	itsDate = itsTime = NULLString;				// date + time
	}

	// set new size and type and allocate pixbuf memory
	//Resize (header.columns, header.rows, IsFloat(), FALSE);
   if (header.imagetype==40)
      return IMREAD_ERR_FORMAT;
      
   theNF = ((itsVersion>=VER_EXTHEADER && itsVersion<100) ? header.f_dim : 1);
   theNY = (header.rows==-1 ? header.longRows : header.rows);
	theNX = (header.columns==-1 ? header.longColumns : header.columns);
	theNZ = ((itsVersion>=VER_VOLUME_BUFFER && itsVersion<100) ? header.longZDim : 1);

	if ( header.imagetype == IMAGE_SPARSE_WORD || header.imagetype == IMAGE_SPARSE_FLOAT )
      return IMREAD_ERR_FORMAT;

	if (itsVersion<VER_VOLUME_BUFFER || itsVersion>=100)
	{
		theNY /= theNF;
		CreateBuffer( myBuffer, theNX, theNY, theNZ, theNF, header.imagetype==IMAGE_FLOAT, header.vector_grid, header.image_sub_type );
	}
	else
	{
		CreateBuffer( myBuffer, theNX,theNY,theNZ,theNF, header.imagetype==IMAGE_FLOAT, header.vector_grid, header.image_sub_type );
	}

	if (header.imagetype == IMAGE_IMX)				// compressed (IMX) file ?
	{
		ImReadError_t err = SCPackOldIMX_Read(theFile,myBuffer);
		if (err!=IMREAD_ERR_NO)
		{
			fclose(theFile);
			return err;
		}
	}			
	else    // no compression
	{
		for (int row=0; row<myBuffer->totalLines; row++)
		{
			if (myBuffer->isFloat)
				fread( &myBuffer->floatArray[row*myBuffer->nx], sizeof(float), myBuffer->nx, theFile );
			else
				fread( &myBuffer->wordArray[row*myBuffer->nx], sizeof(Word), myBuffer->nx, theFile );
			if (ferror(theFile))
				goto errexit;
		}
	}

   if (myList)
	{
      ReadImgExtHeader(theFile,myList);
   }

   fclose(theFile);
   return IMREAD_ERR_NO;

errexit:
   fclose(theFile);
   return IMREAD_ERR_DATA;
}

typedef enum 
	{
      IEH_END = 0,
      IEH_SCALE_X,
      IEH_SCALE_Y,
      IEH_SCALE_Z,
      IEH_SCALE_I,
      IEH_COMMENT,
      IEH_ATTRIBUTE,
      IEH_SCALE_F,
      IEH_ATTRIBUTE_FLOATARRAY,
      IEH_ATTRIBUTE_INTARRAY,
      IEH_ATTRIBUTE_WORDARRAY,
		IEH_TIME,						// long time with milliseconds, TL 24.10.00
		// jetzt duerfen neue Attribute folgen, die bisherigen Daten versteht auch DaVis 6
		IEH_DATE,
		IEH_MAX = IEH_DATE			// max index of this list
   } type_extheader;

typedef struct
   {  type_extheader type;
      int            size;
   } image_extheader;

int SetAttribute( AttributeList** myList, const char* theName, const char* theValue )
{
   AttributeList* ptr = (AttributeList*)malloc(sizeof(AttributeList));
   ptr->name = (char*)malloc(strlen(theName)+1);
   strcpy(ptr->name,theName);
   ptr->value = (char*)malloc(strlen(theValue)+1);
   strcpy(ptr->value,theValue);
   ptr->next = *myList;
   *myList = ptr;
   return 1;
}

void ReplaceNull( char* data, int size )
{
	int i;
	size--;
	for (i=0; i<size; i++)
		if (data[i]=='\0')
			data[i] = '\n';
}

int ReadImgExtHeader( FILE* theFile, AttributeList** myList )
{
   image_extheader item;

   while (fread((char*)&item,1,sizeof(item),theFile))
	{
      char* data = NULL;
      if (item.size>0) {
         data = (char*)malloc(item.size+1);
         data[item.size] = 0; // final 0-byte for strings
         if (!fread(data,1,item.size,theFile))
            return -1;  // "extended header: no tag data"
      }
		if (data)
		{
			switch (item.type) {
			case IEH_END:
				break;
			case IEH_SCALE_X:
				ReplaceNull(data,item.size);
				SetAttribute(myList,"_SCALE_X",data);
				break;
			case IEH_SCALE_Y:
				ReplaceNull(data,item.size);
				SetAttribute(myList,"_SCALE_Y",data);
				break;
			case IEH_SCALE_Z:
				ReplaceNull(data,item.size);
				SetAttribute(myList,"_SCALE_Z",data);
				break;
			case IEH_SCALE_I:
				ReplaceNull(data,item.size);
				SetAttribute(myList,"_SCALE_I",data);
				break;
			case IEH_SCALE_F:
				ReplaceNull(data,item.size);
				SetAttribute(myList,"_SCALE_F",data);
				break;
			case IEH_COMMENT:
				SetAttribute(myList,"_COMMENT",data);
				break;
			case IEH_TIME:
				SetAttribute(myList,"_TIME",data);
				break;
			case IEH_DATE:
				SetAttribute(myList,"_DATE",data);
				break;
			case IEH_ATTRIBUTE:
            {  char* value_pos = strchr(data,'=');
               if (value_pos)
					{
                  *value_pos = '\0';
                  SetAttribute( myList, data, value_pos+1 );
               }
               break;
            }
         }
			free(data);
		}
   }
   return 0;
}

